@@ -405,44 +405,255 @@ class BigQueryExecutor:
         return query_job
 
 
-    def create_partition_table(self,
-                               table_id,
-                               dataset_id=bq_default_dataset(),
-                               project_id=bq_default_project(),
-                               sql=None,
-                               file=None,
-                               use_legacy_sql=False,
-                               write_disposition='WRITE_TRUNCATE',
-                               partition_dates=None,
-                               partition_field='_PARTITIONTIME',
-                               clustering=None,
-                               priority="INTERACTIVE",
-                               schema_path="",
-                               description=None,
-                               **kwargs
-                              ):
+    def delete_backfill_partitions(self,
+                                   backfill_dataset_id,
+                                   table_id,
+                                   partition_field,
+                                   from_partition,
+                                   to_partition
+                                   ):
         """
-        Partition to be generated are either passed through partition_dates or automatically generated using existing partitions.
-        To filter on a specific partition, the filter DATE(_PARTITIONTIME) = {partition_date} can be used in your sql query.
-        """
-        if sql is None and file is None:
-            raise BigQueryExecutorError("Either SQL or file containing the SQL must be provided")
-        if sql is None:
-            sql = read_sql(file, **kwargs)
+        Delete partitions from table in backfill dataset between from_partition and to_partition
 
-        if schema_path != "":
-            self.initiate_table(
-                project_id=project_id,
+        Args:
+            backfill_dataset_id (string): dataset where the backfill tables are created
+            table_id (string): table in backfill dataset that the partitions will be deleted from
+            partition_field (string): the time-unit column that the table is partitioned by
+            from_partition (date): date of the first partition to be deleted
+            to_partition (date): date of the latest partition to be deleted
+
+        Returns:
+            None
+        """
+        client = bigquery.Client()
+
+        sql = f"""
+               DELETE
+                 `{backfill_dataset_id}.{table_id}`
+               WHERE
+                 {partition_field} BETWEEN
+                   CAST('{from_partition}' AS DATE) AND CAST('{to_partition}' AS DATE)
+               """
+
+        query_job = client.query(
+            sql
+        )
+
+        query_job.result() # wait for job to complete
+
+
+    def run_full_backfill(self,
+                          dataset_id,
+                          backfill_dataset_id,
+                          table_id,
+                          schema_path,
+                          sql,
+                          partition_field
+                          ):
+        """
+        Run a full backfill for a time-unit column partitioned table
+
+        Args:
+            dataset_id (string): dataset of the table to be backfilled
+            backfill_dataset_id (string): dataset where the backfill tables are created
+            table_id (string): name of the table to be backfilled
+            schema_path (string): file path to the schema of the table to be backfilled
+            sql (string): query that's used in the pipeline to create this table
+                Note: it has to contain the clause:
+                WHERE <partition_field> BETWEEN CAST(<from_partition> AS DATE) AND CAST(<to_partition> AS DATE)
+            partition_field (string): the time-unit column that the table is partitioned by
+
+        Returns:
+            None
+        """
+        logging.info("Starting full backfill")
+
+        logging.info("Creating backfill dataset if it doesn't exist")
+        self.create_dataset(
+            dataset_id=backfill_dataset_id
+        )
+
+        logging.info("Creating table in backfill dataset with all partitions")
+        self.create_table(
+            dataset_id=backfill_dataset_id,
+            table_id=table_id,
+            schema_path=schema_path,
+            sql=sql,
+            partition=True,
+            partition_field=partition_field,
+            write_disposition="WRITE_TRUNCATE"
+        )
+
+        logging.info("Replacing prod table with backfill table")
+        self.copy_table(
+            source_dataset_id=backfill_dataset_id,
+            source_table_id=table_id,
+            dest_dataset_id=dataset_id,
+            dest_table_id=table_id,
+            write_disposition="WRITE_TRUNCATE"
+        )
+
+        logging.info("Deleting backfill table")
+        self.delete_table(
+            dataset_id=backfill_dataset_id,
+            table_id=table_id
+        )
+
+
+    def run_partial_backfill(self,
+                             dataset_id,
+                             backfill_dataset_id,
+                             table_id,
+                             schema_path,
+                             sql,
+                             partition_field,
+                             from_partition,
+                             to_partition
+                             ):
+        """
+        Run a partial backfill for a time-unit column partitioned table between the dates
+        from_partition and to_partition
+
+        Args:
+            dataset_id (string): dataset of the table to be backfilled
+            backfill_dataset_id (string): dataset where the backfill tables are created
+            table_id (string): name of the table to be backfilled
+            schema_path (string): file path to the schema of the table to be backfilled
+            sql (string): query that's used in the pipeline to create this table
+                Note: it has to contain the clause:
+                WHERE <partition_field> BETWEEN CAST(<from_partition> AS DATE) AND CAST(<to_partition> AS DATE)
+            partition_field (string): the time-unit column that the table is partitioned by
+            from_partition (date): date of the first partition to be backfilled
+            to_partition (date): date of the latest partition to be backfilled
+
+        Returns:
+            None
+        """
+        logging.info("Starting partial backfill")
+
+        logging.info("Creating backfill dataset if it doesn't exist")
+        self.create_dataset(
+            dataset_id=backfill_dataset_id
+        )
+
+        logging.info("Copying table to backfill dataset")
+        self.copy_table(
+            source_dataset_id=dataset_id,
+            source_table_id=table_id,
+            dest_dataset_id=backfill_dataset_id,
+            dest_table_id=table_id,
+            write_disposition="WRITE_TRUNCATE"
+        )
+
+        logging.info("Deleting partitions to be backfilled")
+        self.delete_backfill_partitions(
+            backfill_dataset_id,
+            table_id,
+            partition_field,
+            from_partition,
+            to_partition
+        )
+
+        logging.info("Appending new backfill partitions")
+        self.create_table(
+            dataset_id=backfill_dataset_id,
+            table_id=table_id,
+            schema_path=schema_path,
+            sql=sql,
+            partition=True,
+            partition_field=partition_field,
+            write_disposition="WRITE_APPEND"
+        )
+
+        logging.info("Replacing prod table with backfill table")
+        self.copy_table(
+            source_dataset_id=backfill_dataset_id,
+            source_table_id=table_id,
+            dest_dataset_id=dataset_id,
+            dest_table_id=table_id,
+            write_disposition="WRITE_TRUNCATE"
+        )
+
+        logging.info("Deleting backfill table")
+        self.delete_table(
+            dataset_id=backfill_dataset_id,
+            table_id=table_id
+        )
+
+
+    def run_backfill(self,
+                     backfill,
+                     dataset_id,
+                     table_id,
+                     schema_path,
+                     sql,
+                     partition_field,
+                     from_partition,
+                     to_partition
+                     ):
+        """
+        Run a full or partial backfill for a time-unit column partitioned table
+
+        Args:
+            backfill (string): specifies full or partial backfill (if running a backfill)
+            dataset_id (string): dataset of the table to be backfilled
+            table_id (string): name of the table to be backfilled
+            schema_path (string): file path to the schema of the table to be backfilled
+            sql (string): query that's used in the pipeline to create this table
+                Note: it has to contain the clause:
+                WHERE <partition_field> BETWEEN CAST(<from_partition> AS DATE) AND CAST(<to_partition> AS DATE)
+            partition_field (string): the time-unit column that the table is partitioned by
+
+        Returns:
+            None
+        """
+        logging.info("Starting backfill")
+
+        backfill_dataset = "backfill"
+
+        if backfill not in ("full", "partial"):
+            raise BigQueryExecutorError("BACKFILL env var needs to be 'full' or 'partial'")
+
+        if backfill == "full":
+            self.run_full_backfill(
                 dataset_id=dataset_id,
+                backfill_dataset_id=backfill_dataset,
                 table_id=table_id,
                 schema_path=schema_path,
-                partition=True,
-                partition_field=partition_field,
-                clustering=clustering
+                sql=sql,
+                partition_field=partition_field
             )
 
-        if not self.table_exists(dataset_id=dataset_id, table_id=table_id,project_id=project_id):
-            raise BigQueryExecutorError("To create a partition, please initiate the table first using initiate_table.")
+        elif backfill == "partial":
+            self.run_partial_backfill(
+                dataset_id=dataset_id,
+                backfill_dataset_id=backfill_dataset,
+                table_id=table_id,
+                schema_path=schema_path,
+                sql=sql,
+                partition_field=partition_field,
+                from_partition=from_partition,
+                to_partition=to_partition
+            )
+
+    def create_partitions_legacy(self,
+                                 sql,
+                                 file,
+                                 partition_dates,
+                                 table_id,
+                                 project_id,
+                                 dataset_id,
+                                 bq_start_date,
+                                 bq_end_date,
+                                 write_disposition,
+                                 use_legacy_sql,
+                                 partition_field,
+                                 clustering,
+                                 priority,
+                                 **kwargs
+                                 ):
+        if sql is None:
+            sql = read_sql(file, **kwargs)
 
         if partition_dates is None:
             existing_dates = self.get_existing_partition_dates(
@@ -451,8 +662,8 @@ class BigQueryExecutor:
                 dataset_id=dataset_id
             )
             dates = self.get_partition_dates(
-                start_date=bq_start_date(),
-                end_date=bq_end_date(),
+                start_date=bq_start_date,
+                end_date=bq_end_date,
                 existing_dates=existing_dates
             )
         else:
@@ -486,9 +697,126 @@ class BigQueryExecutor:
             for job in jobs:
                 job.result()
 
+
+    def create_partition_table(self,
+                               table_id,
+                               dataset_id=bq_default_dataset(),
+                               project_id=bq_default_project(),
+                               bq_start_date=bq_start_date(),
+                               bq_end_date=bq_end_date(),
+                               sql=None,
+                               file=None,
+                               use_legacy_sql=False,
+                               write_disposition='WRITE_TRUNCATE',
+                               partition_dates=None,
+                               pipeline_partition_date=None,
+                               partition_field='_PARTITIONTIME',
+                               clustering=None,
+                               priority="INTERACTIVE",
+                               schema_path="",
+                               description=None,
+                               **kwargs
+                              ):
+        """
+        New and legacy ways of creating partition tables:
+
+        Legacy:
+        Partition to be generated are either passed through partition_dates or automatically generated using existing partitions.
+        To filter on a specific partition, the filter DATE(_PARTITIONTIME) = {partition_date} can be used in your sql query.
+
+        New:
+        pipeline_partition_date is required for a normal pipeline run and is a single date.
+        To run a backfill, the BACKFILL env var needs to be set to 'full' or 'partial' and an optional date range 
+        can be set for a partial backfill by setting the FROM_PARTITION and TO_PARTITION env vars.
+        """
+        if sql is None and file is None:
+            raise BigQueryExecutorError("Either SQL or file containing the SQL must be provided")
+
+        if schema_path != "":
+            self.initiate_table(
+                project_id=project_id,
+                dataset_id=dataset_id,
+                table_id=table_id,
+                schema_path=schema_path,
+                partition=True,
+                partition_field=partition_field,
+                clustering=clustering
+            )
+
+        if not self.table_exists(dataset_id=dataset_id, table_id=table_id,project_id=project_id):
+            raise BigQueryExecutorError("To create a partition, please initiate the table first using initiate_table.")
+
+        if "BACKFILL" not in os.environ:
+            if pipeline_partition_date is None: # legacy approach
+                self.create_partitions_legacy(
+                    sql=sql,
+                    file=file,
+                    partition_dates=partition_dates,
+                    table_id=table_id,
+                    project_id=project_id,
+                    dataset_id=dataset_id,
+                    bq_start_date=bq_start_date,
+                    bq_end_date=bq_end_date,
+                    write_disposition=write_disposition,
+                    use_legacy_sql=use_legacy_sql,
+                    partition_field=partition_field,
+                    clustering=clustering,
+                    priority=priority,
+                    **kwargs
+                )
+
+            else: # new approach, no backfill
+                from_partition = pipeline_partition_date
+                to_partition = pipeline_partition_date
+                sql = read_sql(
+                    file=file,
+                    from_partition=from_partition,
+                    to_partition=to_partition
+                )
+
+                # change date format from YYYY-MM-DD to YYYYMMDD for partition_id 
+                partition_id = pipeline_partition_date.replace("-", "")
+                partition_name = self.set_partition_name(
+                    table=table_id,
+                    date=partition_id
+                )
+
+                logging.info("Updating partition: %s", partition_name)
+                self.create_table(
+                    dataset_id=dataset_id,
+                    table_id=partition_name,
+                    schema_path=schema_path,
+                    sql=sql,
+                    partition=True,
+                    partition_field=partition_field
+                )
+
+        else: # new approach, run backfill
+            from_partition = os.environ.get("FROM_PARTITION", bq_start_date)
+            to_partition = os.environ.get("TO_PARTITION", bq_end_date)
+            sql = read_sql(
+                file=file,
+                from_partition=from_partition,
+                to_partition=to_partition
+            )
+            self.run_backfill(
+                backfill=os.environ["BACKFILL"],
+                dataset_id=dataset_id,
+                table_id=table_id,
+                schema_path=schema_path,
+                sql=sql,
+                partition_field=partition_field,
+                from_partition=from_partition,
+                to_partition=to_partition
+            )
+
         if description:
-            self.update_table_description(table_id=table_id, description=description,
-                                          project_id=project_id, dataset_id=dataset_id)
+            self.update_table_description(
+                project_id=project_id,
+                dataset_id=dataset_id,
+                table_id=table_id,
+                description=description
+            )
 
 
     def apply_partition_filter(self, sql, date):
